@@ -2,6 +2,7 @@ import inspect
 import traceback
 from pathlib import Path
 
+import dash
 import dash_ace_persistent
 import dash_core_components as dcc
 import dash_dangerously_set_inner_html
@@ -11,7 +12,7 @@ import numpy as np
 import pandas as pd
 import plotly.express as px
 from dash import Dash
-from dash.dependencies import Input, Output, State
+from dash.dependencies import ALL, Input, Output, State
 from dash_interactive_graphviz import DashInteractiveGraphviz
 from dash_split_pane import DashSplitPane
 from dash_treebeard import DashTreebeard
@@ -21,6 +22,7 @@ from fn_graph.calculation import (
     get_execution_instructions,
 )
 from fn_graph.profiler import Profiler
+from fn_graph import Composer
 from pygments import highlight
 from pygments.formatters import HtmlFormatter
 from pygments.lexers import PythonLexer
@@ -31,12 +33,13 @@ from .parameter_editor import (
     parameter_widgets,
 )
 from .result_renderers import add_default_renderers
+from .layout_helpers import Pane, VStack, HStack, Fill, Scroll
 
 __package__ = "fn_graph_studio"
 
 
 # Load up embedded styles
-# We embed the styles directly in the template ofr portabilities sake
+# We embed the styles directly in the template for portabilities sake
 # I feel there is likely a better way to do this but I cannot fnd it.
 with open(Path(__file__).parent / "styles.css") as f:
     styles = f.read()
@@ -45,60 +48,23 @@ with open(Path(__file__).parent / "highlight.css") as f:
     highlight_styles = f.read()
 
 
-def BasePane(default_style):
-    def wrapper(*args, style=None, **kwargs):
-        return html.Div(*args, style={**default_style, **(style or {})}, **kwargs)
-
-    return wrapper
-
-
-grid_border = 0
-
-Pane = BasePane(dict(position="relative", border=f"{grid_border}px solid black"))
-
-Scroll = BasePane(
-    dict(
-        overflow="auto",
-        width="100%",
-        height="100%",
-        position="absolute",
-        border=f"{grid_border}px solid yellow",
-    )
-)
-Fill = BasePane(
-    dict(
-        height="100%",
-        width="100%",
-        position="absolute",
-        overflow="hidden",
-        border=f"{grid_border}px solid green",
-    )
-)
-HStack = BasePane(
-    dict(
-        display="flex",
-        overflow="hidden",
-        position="relative",
-        border=f"{grid_border}px solid blue",
-    )
-)
-VStack = BasePane(
-    dict(
-        display="flex",
-        flexDirection="column",
-        overflow="hidden",
-        position="relative",
-        border=f"{grid_border}px solid red",
-    )
-)
-
-
 class BaseStudio:
-    def __init__(self, app, composer, renderers=None):
+    def __init__(
+        self,
+        app,
+        *,
+        get_composer,
+        title="Fn Graph Studio",
+        show_profiler=True,
+        editable_parameters=True,
+        renderers=None,
+    ):
+        self._get_composer = get_composer
+        self.show_profiler = show_profiler
+        self.editable_parameters = editable_parameters
+        app.title = title
 
-        app.title = "Fn Graph Studio"
-
-        app.layout = self.layout(composer)
+        app.layout = self.layout()
 
         app.index_string = (
             """
@@ -129,56 +95,66 @@ class BaseStudio:
         """
         )
 
-        parameter_keys = list(get_variable_parameter_keys(composer.parameters()))
-        parameter_ids = list(get_variable_parameter_ids(composer.parameters()))
-
         @app.callback(
             [
-                Output(
-                    component_id="result-function-name", component_property="children"
-                ),
-                Output(component_id="result-type", component_property="children"),
-                Output(component_id="error-container", component_property="children"),
-                Output(component_id="result-container", component_property="children"),
+                Output("result-function-name", "children"),
+                Output("result-type", "children"),
+                Output("error-container", "children"),
+                Output("result-container", "children"),
+                Output("cache-invalidation-store", "data"),
             ],
             [
-                Input(component_id="function-tree", component_property="selected"),
-                Input(component_id="result-processor", component_property="value"),
-                Input(component_id="result-or-definition", component_property="value"),
-                *[
-                    Input(component_id=id_, component_property="value")
-                    for id_ in parameter_ids
-                ],
+                Input("function-tree", "selected"),
+                Input("result-processor", "value"),
+                Input("result-or-definition", "value"),
+                Input("invalidate-cache", "n_clicks"),
+                Input("url", "pathname"),
+                Input({"type": "parameter", "key": ALL}, "value"),
             ],
+            [State("cache-invalidation-store", "data")],
         )
         def populate_result_with_composer(
-            function_name, result_processor, result_or_definition, *parameter_values
+            function_name,
+            result_processor,
+            result_or_definition,
+            invalidate_cache_clicks,
+            path,
+            parameter_values,
+            cache_invalidation_store,
         ):
+            composer = self.get_composer(path)
+            parameters = {
+                input["id"]["key"]: input["value"]
+                for input in dash.callback_context.inputs_list[-1]
+            }
+
+            invalidate_cache = (cache_invalidation_store or 0) < (
+                invalidate_cache_clicks or 0
+            )
+            if invalidate_cache:
+                composer.cache_invalidate(function_name)
+                cache_invalidation_store = invalidate_cache_clicks
+
             return self.populate_result_pane(
                 composer,
                 add_default_renderers(renderers),
-                parameter_keys,
+                parameters,
                 function_name,
                 result_processor,
                 result_or_definition,
-                parameter_values,
-            )
+            ) + (cache_invalidation_store,)
 
         @app.callback(
-            Output(component_id="graphviz-viewer", component_property="dot_source"),
+            Output("graphviz-viewer", "dot_source"),
             [
-                Input(component_id="node-name-filter", component_property="value"),
-                Input(component_id="graph-display-options", component_property="value"),
-                Input(component_id="graph-neighbourhood", component_property="value"),
-                Input(
-                    component_id="graph-neighbourhood-size", component_property="value"
-                ),
-                Input(component_id="function-tree", component_property="selected"),
-                *[
-                    Input(component_id=id_, component_property="value")
-                    for id_ in parameter_ids
-                ],
+                Input("node-name-filter", "value"),
+                Input("graph-display-options", "value"),
+                Input("graph-neighbourhood", "value"),
+                Input("graph-neighbourhood-size", "value"),
+                Input("function-tree", "selected"),
+                Input("url", "pathname"),
             ],
+            [State("cache-invalidation-store", "data")],
         )
         def populate_graph_with_composer(
             node_name_filter,
@@ -186,8 +162,10 @@ class BaseStudio:
             graph_neighbourhood,
             graph_neighbourhood_size,
             selected_node,
-            *parameter_values,
+            url,
+            cache_invalidation_store,
         ):
+            composer = self.get_composer(url)
             return self.populate_graph(
                 composer,
                 node_name_filter,
@@ -195,15 +173,14 @@ class BaseStudio:
                 graph_neighbourhood,
                 graph_neighbourhood_size,
                 selected_node,
-                parameter_keys,
-                parameter_values,
             )
 
         @app.callback(
-            Output(component_id="function-tree", component_property="data"),
-            [Input(component_id="node-name-filter", component_property="value")],
+            Output("function-tree", "data"),
+            [Input("node-name-filter", "value"), Input("url", "pathname")],
         )
-        def populate_tree_with_composer(node_name_filter):
+        def populate_tree_with_composer(node_name_filter, url):
+            composer = self.get_composer(url)
             if node_name_filter:
                 matching_nodes = [
                     key
@@ -226,13 +203,22 @@ class BaseStudio:
 
             return format_tree("_root_", tree)
 
-        sidebar_components = self.sidebar_components(composer)
+        @app.callback(
+            Output("parameters-holder", "children"),
+            [Input("url", "pathname")],
+            [State("parameter_store", "data")],
+        )
+        def populate_parameters_with_composer(url, store):
+            composer = self.get_composer(url)
+            return parameter_widgets(
+                composer.parameters(), store or {}, self.editable_parameters
+            )
+
+        sidebar_components = self.sidebar_components()
 
         @app.callback(
-            [
-                Output(component.id, "style")
-                for component in sidebar_components.values()
-            ],
+            [Output(component.id, "style") for component in sidebar_components.values()]
+            + [Output("node-name-filter", "style")],
             [Input("explorer-selector", "value")],
         )
         def toggle_explorer(explorer):
@@ -245,14 +231,32 @@ class BaseStudio:
                     display="block" if option == explorer else "none",
                 )
                 for option in sidebar_components.keys()
+            ] + [
+                dict(
+                    border="1px solid lightgrey",
+                    width=" calc(100% - 10px)",
+                    margin="5px",
+                    display="block" if "parameters" != explorer else "none",
+                )
             ]
 
         @app.callback(
-            Output("session", "data"),
-            [Input("function-tree", "selected")],
-            [State("session", "data")],
+            Output("parameter_store", "data"),
+            [Input({"type": "parameter", "key": ALL}, "value")],
         )
-        def on_save_to_session(selected, data):
+        def save_parameters_to_session(parameter_values):
+            parameters = {
+                input["id"]["key"]: input["value"]
+                for input in dash.callback_context.inputs_list[-1]
+            }
+            return parameters
+
+        @app.callback(
+            Output("tree_store", "data"),
+            [Input("function-tree", "selected")],
+            [State("tree_store", "data")],
+        )
+        def save_tree_to_session(selected, data):
             if selected and isinstance(selected[0], list):
                 selected = selected[0]
 
@@ -263,7 +267,7 @@ class BaseStudio:
         @app.callback(
             Output("function-tree", "selected"),
             [Input("graphviz-viewer", "selected"), Input("url", "pathname")],
-            [State("session", "data")],
+            [State("tree_store", "data")],
         )
         def update_tree_node(selected, _url, data):
             data = data or {}
@@ -272,13 +276,23 @@ class BaseStudio:
             else:
                 return data.get("selected", "")
 
-    def layout(self, composer):
+    def get_composer(self, path) -> Composer:
+        """
+        Lazily calls the get_composer function based on the path
+
+        This allows it to dynamically choose  a composer.
+        """
+        return self._get_composer(path)
+
+    def layout(self):
         return Pane(
             children=[
                 dcc.Location(id="url", refresh=False),
-                dcc.Store(id="session", storage_type="session"),
+                dcc.Store(id="parameter_store", storage_type="session"),
+                dcc.Store(id="tree_store", storage_type="session"),
+                dcc.Store(id="cache-invalidation-store", storage_type="memory"),
                 DashSplitPane(
-                    [self.sidebar(composer), self.results_pane()],
+                    [self.sidebar_layout(), self.results_pane_layout()],
                     size=400,
                     persistence=True,
                 ),
@@ -286,7 +300,7 @@ class BaseStudio:
             style=dict(width="100%", height="100%", position="absolute"),
         )
 
-    def function_tree(self, component_id, composer):
+    def function_tree(self, component_id):
         return Scroll(
             DashTreebeard(
                 id=component_id,
@@ -355,10 +369,7 @@ class BaseStudio:
             ),
         )
 
-    def parameters(self, composer):
-        return parameter_widgets(composer.parameters())
-
-    def sidebar_components(self, composer):
+    def sidebar_components(self):
         return {
             "graph": Fill(
                 self.function_graph(),
@@ -366,15 +377,11 @@ class BaseStudio:
                 style=dict(display="none"),
             ),
             "tree": Fill(
-                self.function_tree("function-tree", composer),
+                self.function_tree("function-tree"),
                 id="function-tree-holder",
                 style=dict(display="none"),
             ),
-            "parameters": Fill(
-                self.parameters(composer),
-                id="parameters-holder",
-                style=dict(display="none"),
-            ),
+            "parameters": Fill(id="parameters-holder", style=dict(display="none")),
         }
 
     def result_processor(self):
@@ -393,8 +400,8 @@ class BaseStudio:
             debounceChangePeriod=1000,
         )
 
-    def sidebar(self, composer):
-        sidebar_components = self.sidebar_components(composer)
+    def sidebar_layout(self):
+        sidebar_components = self.sidebar_components()
         explorer_options = [
             dict(value=k, label=k.capitalize()) for k in sidebar_components
         ]
@@ -423,12 +430,8 @@ class BaseStudio:
                         type="text",
                         placeholder="Filter nodes",
                         value="",
+                        style=dict(display="none"),
                         persistence=True,
-                        style=dict(
-                            border="1px solid lightgrey",
-                            width=" calc(100% - 10px)",
-                            margin="5px",
-                        ),
                         debounce=True,
                     )
                 ),
@@ -446,7 +449,14 @@ class BaseStudio:
             style=dict(flexShrink=0, height="100%", borderRight="1px solid lightgrey"),
         )
 
-    def results_pane(self):
+    def results_pane_layout(self):
+
+        options = [
+            {"label": "Result", "value": "result"},
+            {"label": "Definition", "value": "definition"},
+        ]
+        if self.show_profiler:
+            options.append({"label": "Profiler", "value": "profiler"})
 
         status_bar = html.Div(
             [
@@ -458,16 +468,18 @@ class BaseStudio:
                         html.Span(id="result-type"),
                     ]
                 ),
-                dcc.RadioItems(
-                    id="result-or-definition",
-                    options=[
-                        {"label": "Result", "value": "result"},
-                        {"label": "Definition", "value": "definition"},
-                        {"label": "Profiler", "value": "profiler"},
+                html.Span(
+                    [
+                        html.Button("Invalidate Cache", id="invalidate-cache"),
+                        dcc.RadioItems(
+                            id="result-or-definition",
+                            options=options,
+                            value="result",
+                            persistence=True,
+                            inputStyle=dict(marginLeft="10px", marginRight="2px"),
+                        ),
                     ],
-                    value="result",
-                    persistence=True,
-                    inputStyle=dict(marginLeft="10px", marginRight="2px"),
+                    style=dict(display="flex", alignItems="center"),
                 ),
             ],
             style=dict(
@@ -479,14 +491,13 @@ class BaseStudio:
 
         result = dcc.Loading(
             Pane(Scroll(id="result-container"), style=dict(height="100%")),
-            style=dict(flexGrow=1),
-            className="result-loader",
             color="#7dc242",
         )
 
         return VStack(
             [status_bar, error_container, result],
             style=dict(flexGrow=1, flexShrink=1, height="100%"),
+            className="result-layout",
         )
 
     def process_result(self, result, result_processor_value):
@@ -698,11 +709,10 @@ class BaseStudio:
         self,
         composer,
         renderers,
-        parameter_keys,
+        parameters,
         function_name,
         result_processor,
         result_or_definition,
-        parameter_values,
     ):
 
         if function_name not in set(composer.dag().nodes()):
@@ -710,20 +720,18 @@ class BaseStudio:
 
         if result_or_definition == "result":
             return self.populate_result(
-                composer,
-                renderers,
-                function_name,
-                result_processor,
-                dict(zip(parameter_keys, parameter_values)),
+                composer, renderers, function_name, result_processor, parameters
             )
         elif result_or_definition == "definition":
             return self.populate_definition(composer, function_name)
         else:
-            return self.populate_profiler(
-                composer, function_name, dict(zip(parameter_keys, parameter_values))
-            )
+            return self.populate_profiler(composer, function_name, parameters)
 
     def update_composer_parameters(self, composer, parameters):
+        """
+        Ensures that boolean parameters et cast correctly
+        """
+
         def smartish_cast(type_, value):
             if issubclass(type_, bool) and isinstance(value, str):
                 return value.lower()[0] == "t"
@@ -745,8 +753,6 @@ class BaseStudio:
         graph_neighbourhood,
         graph_neighbourhood_size,
         selected_node,
-        parameter_keys,
-        parameter_values,
     ):
         graph_display_options = graph_display_options or []
         hide_parameters = "parameters" not in graph_display_options
@@ -754,10 +760,6 @@ class BaseStudio:
         caching = "caching" in graph_display_options
         expand_links = "links" in graph_display_options
 
-        composer = self.update_composer_parameters(
-            composer, dict(zip(parameter_keys, parameter_values))
-        )
-
         G = composer.dag()
         subgraph = set()
 
@@ -765,7 +767,6 @@ class BaseStudio:
             subgraph.update(G.nodes())
 
         if selected_node in G:
-
             if "ancestors" in graph_neighbourhood:
                 subgraph.update(
                     nx.ego_graph(
@@ -884,6 +885,6 @@ def _run_studio(cls, composer, **kwargs):
     """
     Run a studio of type cls for the given composer.
     """
-    app = Dash(__name__)
-    cls(app, composer, **kwargs)
+    app = Dash(__name__, suppress_callback_exceptions=True)
+    cls(app, get_composer=lambda path: composer, **kwargs)
     app.run_server(debug=True)
