@@ -15,16 +15,16 @@ def get_parameter_attrs(key, type_):
 
     for base_type, attrs in parameter_attrs.items():
         if issubclass(type_, base_type):
-            return dict(**attrs, id=f"parameter_{key}")
+            return dict(**attrs, id={"type": "parameter", "key": key})
 
     return None
 
 
-def create_input(key, type_, value):
+def create_input(key, type_, value, editable):
 
     if issubclass(type_, bool):
         return dcc.RadioItems(
-            id=f"parameter_{key}",
+            id={"type": "parameter", "key": key},
             options=[
                 {"label": "True", "value": "True"},
                 {"label": "False", "value": "False"},
@@ -37,7 +37,7 @@ def create_input(key, type_, value):
 
     attrs = get_parameter_attrs(key, type_)
 
-    if attrs:
+    if attrs and editable:
         return dcc.Input(
             **attrs,
             debounce=True,
@@ -69,11 +69,19 @@ def title(string):
     return string.replace("_", " ").capitalize()
 
 
-def parameter_widgets(parameters):
-    widgets = {
-        key: create_input(key, type_, value)
-        for key, (type_, value) in parameters.items()
-    }
+def parameter_widgets(parameters, stored_values, editable_parameters):
+    if editable_parameters:
+        widgets = {
+            key: create_input(
+                key, type_, stored_values[key] if key in stored_values else value, True
+            )
+            for key, (type_, value) in parameters.items()
+        }
+    else:
+        widgets = {
+            key: create_input(key, type_, value, False)
+            for key, (type_, value) in parameters.items()
+        }
 
     def recursive_tree():
         return defaultdict(recursive_tree)
@@ -94,8 +102,8 @@ def parameter_widgets(parameters):
             if level == 0:
                 return html.Div(
                     [
-                        create_containers(level + 1, ckey, cvalue)
-                        for ckey, cvalue in value.items()
+                        create_containers(level + 1, container_key, container_value)
+                        for container_key, container_value in value.items()
                     ],
                     style=dict(padding="0.5rem"),
                 )
@@ -113,8 +121,10 @@ def parameter_widgets(parameters):
                         ),
                         html.Div(
                             [
-                                create_containers(level + 1, ckey, cvalue)
-                                for ckey, cvalue in value.items()
+                                create_containers(
+                                    level + 1, container_key, container_value
+                                )
+                                for container_key, container_value in value.items()
                             ],
                             style=dict(padding="0.25rem"),
                         ),
@@ -127,8 +137,14 @@ def parameter_widgets(parameters):
                         html.Label(title(key), style=dict(fontWeight="bold")),
                         html.Div(
                             [
-                                html.Div([create_containers(level + 1, ckey, cvalue)])
-                                for ckey, cvalue in value.items()
+                                html.Div(
+                                    [
+                                        create_containers(
+                                            level + 1, container_key, container_value
+                                        )
+                                    ]
+                                )
+                                for container_key, container_value in value.items()
                             ],
                             style=dict(
                                 marginLeft="0.25rem",
